# Update and supplement original data set.
remove(list=ls())

library(pmml)

# ##################
# Start user config.
options("stringsAsFactors"=FALSE)
data_dir <- file.path(getwd(), "data")
out_dir <- file.path(getwd(), "output")

tran_orig_df <- read.csv(file.path(data_dir, "transaction_orig.csv"))

n_tran <- 100000
n_acct <- 120 # [apca] 12nnnnnnn
n_apca <- 12  # nnn-nnn
avg_tx_gap_secs <- 5

min_exec_lag_secs <- 30 * 60
max_exec_lag_secs <- 120 * 60
# End user config.
# ################

set.seed(13)

tx_lags <- cumsum(sample(1:20, n_tran, prob=20:1, replace=TRUE))
tran_df <- data.frame(tx_id=sprintf("TN%09.0f", 1123500 + tx_lags))

apca_lst <- paste(sample(101:999, n_apca, replace=FALSE), "-", 
  sample(101:999, n_apca, replace=FALSE), sep="")
apca_all <- sample(apca_lst, n_acct, replace=TRUE)

acct_df <- data.frame(account_id=paste(apca_all, 123000000+sample(1e5:(1e6-1), 
  n_acct, replace=FALSE)), apca=apca_all)

tran_df[, c("account_id","apca")] <- acct_df[sample(1:n_acct, n_tran, replace=TRUE),]

pc <- table(tran_orig_df[, "postcode"])
tran_df[, "postcode"] <- sample(as.integer(names(pc)), n_tran, replace=TRUE, prob=pc)

curr_df <- read.csv(file.path(data_dir, "cfg_currency.csv"))
tran_df[, "currency"] <- sample(curr_df[, "currency"], n_tran, 
  prob=curr_df[, "freq"], replace=TRUE)

desc_df <- read.csv(file.path(data_dir, "cfg_description.csv"))
idx <- sample(1:nrow(desc_df), n_tran, prob=desc_df[, "freq"], replace=TRUE)
tran_df[, c("description", "merchant_cat_code", "merchant_name")] <- 
  desc_df[idx, c("description", "merchant_cat_code", "merchant_name")]

type_df <- read.csv(file.path(data_dir, "cfg_type.csv"))
tran_df[, "type"] <- sample(type_df[, "type"], n_tran, 
  prob=type_df[, "freq"], replace=TRUE)

# Interval to sample for post times.
tseq <- seq.POSIXt(as.POSIXct("2020-01-01 10:17:33"), by="sec", 
  length.out=avg_tx_gap_secs * n_tran)
hrs <- as.integer(format(tseq, "%H"))
hours_df <- read.csv(file.path(data_dir, "cfg_hour.csv"))
freq <- hours_df[match(hrs, hours_df[, 1]), "freq"]
tran_df[, "postDateTime"] <- sort(sample(tseq, n_tran, prob=freq, replace=TRUE))

lseq <- min_exec_lag_secs:max_exec_lag_secs
tran_df[, "executeDateTime"] <- tran_df[, "postDateTime"] + sample(lseq, n_tran,
  prob=1 + abs(0.5 * (min_exec_lag_secs + max_exec_lag_secs) - lseq), replace=TRUE)

tran_df[, "amount"] <- round(60 * rchisq(n_tran, 3), 2)
tran_df[, "post_hour_of_day"] <- as.integer(format(tran_df[, "postDateTime"], "%H"))
tran_df[, "post_day_of_week"] <- as.integer(format(tran_df[, "postDateTime"], "%w"))

tran_df[, "time_since_last_tx"] <- 0
tran_df[, "tx_last_hour"] <- 0
tran_df[, "tx_last_10_mins"] <- 0
for (acct in acct_df[, "account_id"]) {
  idx <- (tran_df[, "account_id"] == acct)
  xx <- tran_df[idx,]
  diffs <- c(0, diff(xx[, "postDateTime"]))
  tran_df[idx, "time_since_last_tx"] <- diffs
  cs <- cumsum(diffs)
  tran_df[idx, "tx_last_hour"] <- unlist(lapply(1:sum(idx), 
    function(x, cs){sum((cs[x] - cs > 0) & (cs[x] - cs < 3600))}, cs=cs))
  tran_df[idx, "tx_last_10_mins"] <- unlist(lapply(1:sum(idx), 
    function(x, cs){sum((cs[x] - cs > 0) & (cs[x] - cs < 600))}, cs=cs))
}

# Write transactions used for scoring in SB.
write.csv(tran_df, file=file.path(out_dir, "transactions.csv"), 
  row.names=FALSE, na="")

# Perform PCA.
x <- as.matrix(tran_df[, c("amount", "post_hour_of_day", "post_day_of_week",
  "time_since_last_tx", "tx_last_hour", "tx_last_10_mins")])
pcaFit <- princomp(x, cor=TRUE)

# Check that the expression for s2 gives the correct values 
# of PCA scores for each PCA in each row. 
s1 <- pcaFit$scores
s2 <- ((x - t(pcaFit$center)[rep(1, nrow(x)), ]) / 
    (t(pcaFit$scale)[rep(1, nrow(x)), ])) %*% pcaFit$loadings
all.equal(s1, s2)

# Examine the cumulative proportion of variance among components
# to determine which are to be used in the oddity calculation.
# 20201001: use the first 4 components (88%) in computing oddity.
summary(pcaFit)

oddScores <- sqrt(apply(pcaFit$scores[, 1:3]^2, 1, sum))
hist(oddScores)

x1 <- x[1, , drop=FALSE]
x1_box <- xform_wrap(x1)

# Reproduce the "s2" expression to calculate the scores in PMML.
for (n in 1:4) {
  x1_box <- xform_function(x1_box,
    orig_field_name=paste(colnames(x1), collapse=", "),
    new_field_name=paste("score", n, sep=""),
    expression=paste(paste(colnames(x1), pcaFit$loadings[, n] / pcaFit$scale, 
      collapse=" + ", sep="*"), "-", 
      sum(pcaFit$center * pcaFit$loadings[, n] / pcaFit$scale))
  )
}

# Take the sqrt of the sum of squares of the scores for the 
# top several componentsin PMML. This is the oddity score.
x1_box <- xform_function(x1_box,
  orig_field_name=paste("score", 1:4, sep="", collapse=", "),
  new_field_name="oddity",
  expression=paste("(", paste(paste("score", 1:4, " ^ 2", sep=""),
    collapse=" + "), ") ^ .5", sep="")
)

# For now, generate only <LocalTransformations>, the hard part.
# TODO: Eliminate manual step of incorporating <LocalTransformations>
#       into the PMML file.
x1_fit <- lm(amount ~ post_hour_of_day+
    post_day_of_week+time_since_last_tx+tx_last_hour+tx_last_10_mins+score1+
    score2+score3+score4+oddity, 
  data=data.frame(tran_df[, c("amount", "post_hour_of_day", "post_day_of_week",
    "time_since_last_tx", "tx_last_hour", "tx_last_10_mins")], score1=1,
    score2=1, score3=1, score4=1, oddity=1))

loctran_pmml <- pmml(NULL, transforms=x1_box) # <LocalTransformations> only
save_pmml(loctran_pmml, file.path(out_dir, "LocalTransformations.pmml"), version=4.3)
# odd_pmml <- pmml(x1_fit, transforms=x1_box)
# save_pmml(odd_pmml, file.path(out_dir, "oddity.pmml"), version=4.3)

