/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.functions;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.client.CustomFunctionResolver;

public class ConfigurationFunctions {

	private static Tuple configuration;
	
	/**
	 * Sets a tuple as configuration object.
	 * 
	 * @param configuration the configuration tuple
	 * @return the input configuration tuple is returned
	 */
	@CustomFunctionResolver("setConfigurationCustomFunctionResolver0")
	public static Tuple setConfiguration(Tuple configuration) {
		ConfigurationFunctions.configuration = configuration;
		return ConfigurationFunctions.configuration;
	}

	public static CompleteDataType setConfigurationCustomFunctionResolver0(CompleteDataType configuration) {
		return configuration;
	}

	/**
	 * Sets a configuration tuple delta on top of a configuration tuple base. 
	 * 
	 * If the base name exists and it is a non-primitive type (tuple) then the call will recurse through then
	 * hierarchy of primitive types. If the base name exists and is a primiteve type, then it will be updated
	 * with the value from the delta. 
	 * 
	 * If the base name doesn't exist, whether a primitive or complex type, then it will be added to the base 
	 * from the delta.
	 *  
	 * @param base the base configuration tuple
	 * @param delta the delta configuration tuple
	 * @return the merged configuration tuple
	 */
	@CustomFunctionResolver("setConfigurationCustomFunctionResolver1")
	public static Tuple setConfiguration(Tuple base, Tuple delta) {
		ConfigurationFunctions.configuration = base;
		if(delta != null) {
			try {
				processTuple(ConfigurationFunctions.configuration, delta);
			} catch (TupleException e) {
				e.printStackTrace();
			}
		}
		
		return ConfigurationFunctions.configuration;
	}

	public static CompleteDataType setConfigurationCustomFunctionResolver1(CompleteDataType base, CompleteDataType delta) {
		return base;
	}

	/**
	 * Merges a configuration tuple delta on top of a configuration tuple base and returns the merged tuple. 
	 * 
	 * @param base the base configuration tuple
	 * @param delta the delta configuration tuple
	 * @return the merged configuration tuple
	 */
	@CustomFunctionResolver("mergeConfigurationCustomFunctionResolver0")
	public static Tuple mergeConfiguration(Tuple base, Tuple delta) {
		if(delta != null) {
			try {
				processTuple(base, delta);
			} catch (TupleException e) {
				e.printStackTrace();
			}
		}
		
		return base;
	}

	public static CompleteDataType mergeConfigurationCustomFunctionResolver0(CompleteDataType base, CompleteDataType delta) {
		return base;
	}

	/**
	 * Process a base and delta tuple objects.
	 * 
	 * @param base the base configuration
	 * @param delta the delta configuration
	 * @throws TupleException if there is an error
	 */
	private static void processTuple(Tuple base, Tuple delta) throws TupleException {
		if(delta == null) 
			return;
		
		Schema deltaSchema = delta.getSchema();
		for(int idx = 0; idx < deltaSchema.getFieldCount(); idx++) {
			Field f = deltaSchema.getField(idx);
			if(delta.isNull(f.getName()) == true)
				continue;
			else if(base.isNull(f.getName()) == true) {
				base.setField(f.getName(), delta.getField(f.getName()));
			}
			else if(f.getDataType() == DataType.TUPLE){
				processTuple(base.getTuple(f.getName()), delta.getTuple(f.getName()));
			}
			else {
				base.setField(f.getName(), delta.getField(f.getName()));
			}
		}		
	}
	
	/**
	 * Retrieve the configuration tuple.
	 * 
	 * @param schema the schema for the configuration tuple
	 * @return the configuration tuple
	 */
	@CustomFunctionResolver("getConfigurationCustomFunctionResolver0")
	public static Tuple getConfiguration(Tuple schema) {
		return configuration;
	}

	public static CompleteDataType getConfigurationCustomFunctionResolver0(CompleteDataType schema) {
		return schema;
	}

	/**
	 * Splits a configuration into application and context depending on the boolean app flag.
	 * 
	 * @param configuration the configuration
	 * @param dictionary the dictionary
	 * @param app true if application otherwise false
	 * @return the tuple
	 */
	@CustomFunctionResolver("splitConfigurationCustomFunctionResolver0")
	public static Tuple splitConfiguration(Tuple configuration, String dictionary, boolean app) {
		JSONObject dictionaryObj = JSON.parseObject(dictionary).getJSONObject("configuration");
		Tuple target = splitBuildTuple(configuration, dictionaryObj, app);
		
		return target;
	}

	private static Tuple splitBuildTuple(Tuple source, JSONObject dictionary, boolean app) {
		Schema schema = source.getSchema();
		Tuple target = schema.createTuple();
		int fieldCount = 0;
		
		for(int idx = 0; idx < schema.getFieldCount(); idx++) {
			Field f = schema.getField(idx);
			try {
				if(source.isNull(f.getName()) || dictionary.getJSONObject(f.getName()) == null)
					continue;
	
				if(f.getDataType() == DataType.TUPLE) {
					Tuple thisTarget = splitBuildTuple(source.getTuple(f.getName()), 
							dictionary.getJSONObject(f.getName()), app);
					if(thisTarget != null) {
						target.setField(f.getName(), thisTarget);
						fieldCount++;
					}
					else {
						target.setNull(f.getName());
					}
				}
				else {
					JSONObject thisDictionary = dictionary.getJSONObject(f.getName());
					if(thisDictionary.getBooleanValue("$application") == app) {
						target.setField(f.getName(), source.getField(f.getName()));
						fieldCount++;
					}
				}
			}
			catch(Exception e) {
				e.printStackTrace();
			}
		}

		if(fieldCount > 0)
			return target;
		return null;
	}
	
	public static CompleteDataType splitConfigurationCustomFunctionResolver0(CompleteDataType configuration, CompleteDataType dictionary, CompleteDataType app) {
		return configuration;
	}

}
