/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.functions;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.client.CustomFunctionResolver;

public class TupleFunctions {

	private static Schema elementNameValueStringSchema = new Schema(null, Schema.createField(DataType.STRING, "name"), 
			Schema.createField(DataType.STRING, "value"));
	
	/**
	 * Converts a dictionary data structure into a tuple data structure. The dictionary must consist
	 * of a list of tuples with fields name and value. The value must be castable into the data type
	 * in the specified tuplespec tuple. The target tuple will not be used, it is required for Schema 
	 * structure only, however must not be passed as a null tuple because of how custom functions work.
	 * 
	 * @param dictionary list of name-value pairs
	 * @param tuplespec the specification for the output tuple structure
	 * @param dictNameField the field containing the element name
	 * @param dictValueField the field containing the element value
	 * @return the tuple populated with field data from the dictionary
	 */
	@CustomFunctionResolver("dictionaryToTupleResolver0")
	public static Tuple dictionaryToTuple(List<?> dictionary, Tuple tuplespec, String dictNameField, String dictValueField) {
		Schema schema = tuplespec.getSchema();
		Tuple out = schema.createTuple();
		if(dictionary == null) return out;

		// Create a map of dictionary field name to dictionary tuple 
		Map<String,Tuple> dictionaryMap = new HashMap<String,Tuple>();
		for(int idx = 0; idx < dictionary.size(); idx++) {
			Tuple dictionaryTuple = (Tuple) dictionary.get(idx);
			try {
				dictionaryMap.put(dictionaryTuple.getString(dictNameField), dictionaryTuple);
			}
			catch(TupleException e) {
			}
		}
		
		// Loop over the tuple schema fields and populate on the output
		for(int idx = 0; idx < schema.getFieldCount(); idx++) {
			Field f = schema.getField(idx);
			Tuple dictionaryTuple = dictionaryMap.get(f.getName());
			if(dictionaryTuple != null) {
				try {
					// Get the value from the dictionary
					String value = dictionaryTuple.getField(dictValueField).toString();
					
					// Set the field value on the output tuple based on the output field type
					//   Some of these may fail if the type is not consistent
					if(f.getCompleteDataType() == CompleteDataType.forBoolean()) {
						out.setBoolean(f.getName(), Boolean.parseBoolean(value));
					}
					else if(f.getCompleteDataType() == CompleteDataType.forDouble()) {
						out.setDouble(f.getName(), Double.parseDouble(value));
					}
					else if(f.getCompleteDataType() == CompleteDataType.forInt()) {
						out.setInt(f.getName(), Integer.parseInt(value));
					}
					else if(f.getCompleteDataType() == CompleteDataType.forLong()) {
						out.setLong(f.getName(), Long.parseLong(value));
					}
					else if(f.getCompleteDataType() == CompleteDataType.forString()) {
						out.setString(f.getName(), value);
					}
					else if(f.getCompleteDataType() == CompleteDataType.forTimestamp()) {
						out.setTimestamp(f.getName(), Timestamp.fromString(value));
					}
					else {
						// Other types not supported, will be skipped
					}
				}
				catch(StreamBaseException e) {
				}
			}
		}
		
		return out;
	}
	
	/**
	 * Resolver.
	 *  
	 * @param dictionary the dictionary type
	 * @param tuplespec the tuplespec type
	 * @param dictNameField the dictionary name field
	 * @param dictValueField the dictionary value field
	 * @return the data type
	 */
	public static CompleteDataType dictionaryToTupleResolver0(CompleteDataType dictionary, CompleteDataType tuplespec, 
			CompleteDataType dictNameField, CompleteDataType dictValueField) {
		return tuplespec;
	}	

	/**
	 * Converts a dictionary data structure into a tuple data structure. Will use default values of
	 * "name" and "value" for the dictionary elements.
	 * 
	 * @param dictionary list of name-value pairs
	 * @param tuplespec the specification for the output tuple structure
	 * @return the tuple populated with field data from the dictionary
	 */
	@CustomFunctionResolver("dictionaryToTupleResolver1")
	public static Tuple dictionaryToTuple(List<?> dictionary, Tuple tuplespec) {
		return TupleFunctions.dictionaryToTuple(dictionary, tuplespec, "name", "value");
	}
	
	/**
	 * Resolver.
	 *  
	 * @param dictionary the dictionary type
	 * @param tuplespec the tuplespec type
	 * @return the data type
	 */
	public static CompleteDataType dictionaryToTupleResolver1(CompleteDataType dictionary, CompleteDataType tuplespec) {
		return tuplespec;
	}	

	
	
	/**
	 * Converts a tuple into a standard dictionary structure. Will use "name" for and "value" for 
	 * the dictionary elements, and the datatype will be "string".
	 *  
	 * @param tuple the tuple to convert
	 * @return the dictionary tuple
	 */
	@CustomFunctionResolver("tupleToDictionaryResolver0")
	public static List<Tuple> tupleToDictionary(Tuple tuple) {
		if(tuple == null) return null;
		Schema tupleSchema = tuple.getSchema();
		Schema dictSchema = elementNameValueStringSchema;
		
		List<Tuple> output = new ArrayList<Tuple>();
		for(int idx = 0; idx < tupleSchema.getFieldCount(); idx++) {
			Field tupleField = tupleSchema.getField(idx);
			try {
				if(!tuple.isNull(tupleField.getName())) {
					Tuple dictTuple = dictSchema.createTuple();
					dictTuple.setString("name", tupleField.getName());
					dictTuple.setString("value", tuple.getField(tupleField.getName()).toString());
					output.add(dictTuple);
				}
			}
			catch(TupleException e) {
			}
		}
		return output;
	}
	
	/**
	 * Resolver.
	 *  
	 * @param tuple the tuple type
	 * @return the data type
	 */
	public static CompleteDataType tupleToDictionaryResolver0(CompleteDataType tuple) {
		return CompleteDataType.forList(CompleteDataType.forTuple(elementNameValueStringSchema));
	}
	
	
	/**
	 * Converts a tuple into a standard dictionary structure. Will use "name" for and "value" for 
	 * the dictionary elements and use the datatype in the tuplespec.
	 *  
	 * @param tuple the tuple to convert
	 * @param tuplespec the specification for the output dictionary structure
	 * @return the dictionary tuple
	 */
	@CustomFunctionResolver("tupleToDictionaryResolver1")
	public static List<Tuple> tupleToDictionary(Tuple tuple, Tuple tuplespec) {
		if(tuple == null) return null;
		Schema tupleSchema = tuple.getSchema();
		Schema dictSchema = tuplespec.getSchema();
		
		List<Tuple> output = new ArrayList<Tuple>();
		for(int idx = 0; idx < tupleSchema.getFieldCount(); idx++) {
			Field tupleField = tupleSchema.getField(idx);
			try {
				if(!tuple.isNull(tupleField.getName())) {
					String value = tuple.getField(tupleField.getName()).toString();					
					Tuple dictTuple = createDictTuple(tupleField.getName(), dictSchema, "name", "value", value);
					output.add(dictTuple);
				}
			}
			catch(StreamBaseException e) {
				e.printStackTrace();
			}
		}
		return output;
	}
	
	public static CompleteDataType tupleToDictionaryResolver1(CompleteDataType tuple, CompleteDataType tuplespec) {
		return CompleteDataType.forList(tuplespec);
	}
	
	
	
	/**
	 * Converts a tuple into a standard dictionary structure. Will use "name" for and "value" for 
	 * the dictionary elements and use the datatype in the tuplespec.
	 *  
	 * @param tuple the tuple to convert
	 * @param tuplespec the specification for the output dictionary structure
	 * @param dictNameField the dictionary field name
	 * @param dictValueField the dictionary value name
	 * @return the dictionary tuple
	 */
	@CustomFunctionResolver("tupleToDictionaryResolver2")
	public static List<Tuple> tupleToDictionary(Tuple tuple, Tuple tuplespec, String dictNameField, String dictValueField) {
		if(tuple == null) return null;
		Schema tupleSchema = tuple.getSchema();
		Schema dictSchema = tuplespec.getSchema();
		
		List<Tuple> output = new ArrayList<Tuple>();
		for(int idx = 0; idx < tupleSchema.getFieldCount(); idx++) {
			Field tupleField = tupleSchema.getField(idx);
			try {
				if(!tuple.isNull(tupleField.getName())) {
					String value = tuple.getField(tupleField.getName()).toString();					
					Tuple dictTuple = createDictTuple(tupleField.getName(), dictSchema, dictNameField, dictValueField, value);
					output.add(dictTuple);
				}
			}
			catch(StreamBaseException e) {
				e.printStackTrace();
			}
		}
		return output;
	}
	
	/**
	 * Resolver.
	 *  
	 * @param tuple the tuple type
	 * @param tuplespec the tuplespec type
	 * @param dictNameField the dictionary name field
	 * @param dictValueField the dictionary value field
	 * @return the data type
	 */
	public static CompleteDataType tupleToDictionaryResolver2(CompleteDataType tuple, CompleteDataType tuplespec,
			CompleteDataType dictNameField, CompleteDataType dictValueField) {
		return CompleteDataType.forList(tuplespec);
	}
	
	/**
	 * Create a dictionary tuple.
	 * 
	 * @param name dictionary field name
	 * @param dictSchema dictionary schema
	 * @param dictNameField dictionary name field
	 * @param dictValueField dictionary value field
	 * @param value value field
	 * @return the tuple
	 * @throws StreamBaseException on exception
	 */
	private static Tuple createDictTuple(String name, Schema dictSchema, String dictNameField, String dictValueField, String value) throws StreamBaseException {
		Tuple dictTuple = dictSchema.createTuple();
		dictTuple.setString(dictNameField, name);
		Field dictField = dictSchema.getField(dictValueField);
		
		// Set the field value on the output tuple based on the output field type
		//   Some of these may fail if the type is not consistent
		if(dictField.getCompleteDataType() == CompleteDataType.forBoolean()) {
			dictTuple.setBoolean(dictValueField, Boolean.parseBoolean(value));
		}
		else if(dictField.getCompleteDataType() == CompleteDataType.forDouble()) {
			dictTuple.setDouble(dictValueField, Double.parseDouble(value));
		}
		else if(dictField.getCompleteDataType() == CompleteDataType.forInt()) {
			dictTuple.setInt(dictValueField, Integer.parseInt(value));
		}
		else if(dictField.getCompleteDataType() == CompleteDataType.forLong()) {
			dictTuple.setLong(dictValueField, Long.parseLong(value));
		}
		else if(dictField.getCompleteDataType() == CompleteDataType.forString()) {
			dictTuple.setString(dictValueField, value);
		}
		else if(dictField.getCompleteDataType() == CompleteDataType.forTimestamp()) {
			dictTuple.setTimestamp(dictValueField, Timestamp.fromString(value));
		}
		else {
			// Other types not supported, will be skipped
			System.out.println("unknown");
		}
		
		return dictTuple;
	}
	
	
	/**
	 * Retrieves dictional value
	 * @param dictionary the dictionary
	 * @param dictNameField field in the dictionary for name
	 * @param dictValueField field in the dictionary for value
	 * @param dictName the name to match
	 * @return the value
	 */
	@CustomFunctionResolver("getDictionaryValueStringResolver")
	public static String getDictionaryValueString(List<?> dictionary, String dictNameField, String dictValueField, String dictName) {
		if(dictionary == null || dictionary.size() == 0 || dictNameField == null || dictValueField == null || dictName == null)
			return null;
		
		try {
			for(int idx = 0; idx < dictionary.size(); idx++) {
				Tuple tuple = (Tuple) dictionary.get(idx);				
				if(tuple.getSchema().hasField(dictNameField) == true && tuple.getSchema().hasField(dictValueField) == true && 
						tuple.isNull(dictNameField) == false && tuple.getString(dictNameField).equals(dictName)) {
					return tuple.getString(dictValueField);
				}
			}
			
			return null;
		}
		catch(Exception ex) {			
			ex.printStackTrace();
		}

		return null;
	}
	
	/**
	 * Resolver for retrieving dictionary value
	 * @param dictionary tuple type
	 * @param dictNameField tuple type
	 * @param dictValueField tuple type
	 * @param dictName tuple type
	 * @return tuple type
	 */
	public static CompleteDataType getDictionaryValueStringResolver(CompleteDataType dictionary,
			CompleteDataType dictNameField, CompleteDataType dictValueField, CompleteDataType dictName) {
		return CompleteDataType.forString();
	}
	
	
	/**
	 * Retrieves dictional value
	 * @param dictionary the dictionary
	 * @param dictNameField field in the dictionary for name
	 * @param dictValueField field in the dictionary for value
	 * @param dictName the name to match
	 * @return the value
	 */
	@CustomFunctionResolver("getDictionaryValueDoubleResolver")
	public static Double getDictionaryValueDouble(List<?> dictionary, String dictNameField, String dictValueField, String dictName) {
		if(dictionary == null || dictionary.size() == 0 || dictNameField == null || dictValueField == null || dictName == null)
			return null;
		
		try {
			for(int idx = 0; idx < dictionary.size(); idx++) {
				Tuple tuple = (Tuple) dictionary.get(idx);				
				if(tuple.getSchema().hasField(dictNameField) == true && tuple.getSchema().hasField(dictValueField) == true && 
						tuple.isNull(dictNameField) == false && tuple.getString(dictNameField).equals(dictName)) {
					return tuple.getDouble(dictValueField);
				}
			}
			
			return null;
		}
		catch(Exception ex) {			
			ex.printStackTrace();
		}

		return null;
	}
	
	/**
	 * Resolver for retrieving dictionary value
	 * @param dictionary tuple type
	 * @param dictNameField tuple type
	 * @param dictValueField tuple type
	 * @param dictName tuple type
	 * @return tuple type
	 */
	public static CompleteDataType getDictionaryValueDoubleResolver(CompleteDataType dictionary,
			CompleteDataType dictNameField, CompleteDataType dictValueField, CompleteDataType dictName) {
		return CompleteDataType.forDouble();
	}
		
}

