/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.operators.container;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;

import com.streambase.sb.*;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.operator.*;
import com.streambase.sb.sbd.StreamBaseServer;
import com.streambase.sb.sbd.net.xmlrpc.current.ADD_CONTAINERHandler;
import com.streambase.sb.sbd.net.xmlrpc.current.MODIFY_CONTAINERHandler;
import com.streambase.sb.sbd.net.xmlrpc.current.REMOVE_CONTAINERHandler;
import com.streambase.sb.util.Util;

public class ContainerOperator extends Operator implements Parameterizable {

	public static final long serialVersionUID = 1532601011527L;
	private Logger logger;

	// Properties
	private OperationEnum operation;

	// Enum definition for property operation
	public static enum OperationEnum {
		ADD("Add"), MODIFY("Modify"), REMOVE("Remove");

		private final String rep;

		private OperationEnum(String s) {
			rep = s;
		}

		public String toString() {
			return rep;
		}
	}

	private DeliveryEnum delivery;

	// Enum definition for property delivery
	public static enum DeliveryEnum {
		FILE("File"), IN_STREAM_BLOB("In Stream Blob");

		private final String rep;

		private DeliveryEnum(String s) {
			rep = s;
		}

		public String toString() {
			return rep;
		}
	}

	private String containerName;
	private String containerAppFilename;
	private String containerAppBlob;
	private String containerOptions;
	private String displayName = "Container Operator";
	
	// Local variables
	private int inputPorts = 1;
	private int outputPorts = 1;
	private Schema inputSchema;
	private Schema[] outputSchemas;

	private Schema statusSchema = new Schema(null, new Field[] {
            new Field("success", CompleteDataType.forBoolean()),
            new Field("reason", CompleteDataType.forString()),
            new Field("message", CompleteDataType.forString()),
    });
    
    private StreamBaseServer server = null;
	
	public ContainerOperator() {
		super();
		logger = getLogger();
		setPortHints(inputPorts, outputPorts);
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setContainerName("");
		setContainerAppFilename("");
		setContainerOptions("");
	}

	/**
	 * Typecheck the operator
	 * @throws TypecheckException on exception
	 */
	public void typecheck() throws TypecheckException {
		// typecheck: require a specific number of input ports
		requireInputPortCount(inputPorts);
		setCaptureStrategy(CaptureTransformStrategy.NEST);

		// Set input schema
		this.inputSchema = this.getInputSchema(0);
		
		// Check containerName tuple exists
		try {
			Field f = inputSchema.getField(getContainerName());
			if(f.getCompleteDataType() != CompleteDataType.forString()) {
				throw new PropertyTypecheckException("containerName", "Container name tuple must be a string type");
			}
		} catch (TupleException e) {
			throw new PropertyTypecheckException("containerName", "Container name tuple not present in input");
		}

		
		// If ADD and FILE, check containerAppFileName tuple exists
		if(this.getOperation() == OperationEnum.ADD && this.getDelivery() == DeliveryEnum.FILE) {
			try {
				Field f = inputSchema.getField(getContainerAppFilename());
				if(f.getCompleteDataType() != CompleteDataType.forString()) {
					throw new PropertyTypecheckException("containerAppFilename", "Container application file name tuple must be a string type");
				}
			} catch (TupleException e) {
				throw new PropertyTypecheckException("containerAppFilename", "Container application file name tuple not present in input");
			}
		}
		
		// If ADD and BLOB, check containerAppBlob tuple exists
		if(this.getOperation() == OperationEnum.ADD && this.getDelivery() == DeliveryEnum.IN_STREAM_BLOB) {
			try {
				Field f = inputSchema.getField(getContainerAppBlob());
				if(f.getCompleteDataType() != CompleteDataType.forBlob()) {
					throw new PropertyTypecheckException("containerAppBlob", "Container application blob tuple must be a blob type");
				}
			} catch (TupleException e) {
				throw new PropertyTypecheckException("containerAppBlob", "Container application blob tuple not present in input");
			}
		}
		
		// If MODIFY, check containerOptions tuple exists
		if(this.getOperation() == OperationEnum.MODIFY) {
			try {
				Field f = inputSchema.getField(getContainerOptions());
				if(!f.getCompleteDataType().equals(CompleteDataType.forList(CompleteDataType.forString()))) {
					throw new PropertyTypecheckException("containerOptions", "Container options tuple must be a list of string type");
				}
			} catch (TupleException e) {
				throw new PropertyTypecheckException("containerOptions", "Container options tuple not present in input");
			}
		}
		
		List<Field> outputSchemaFields = new ArrayList<Field>();
		for(int idx = 0; idx < inputSchema.getFieldCount(); idx++)
			outputSchemaFields.add(inputSchema.getField(idx));
		outputSchemaFields.add(new Field("status", CompleteDataType.forTuple(this.statusSchema)));
		
		// Set output schema to same as input
		setOutputSchema(0, new Schema(null, outputSchemaFields));
	}

	/**
	 * Process a tuple
	 * @param inputPort input port
	 * @param tuple input tuple
	 * @throws StreamBaseException on exception
	 */
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {

		String containerName = tuple.getString(getContainerName());
		
		if(this.getOperation() == OperationEnum.ADD) {
			if(this.getDelivery() == DeliveryEnum.FILE) {
				String containerAppFilename = tuple.getString(getContainerAppFilename());
				this.processAddContainer(containerName, containerAppFilename, tuple);
			}
			else if(this.getDelivery() == DeliveryEnum.IN_STREAM_BLOB) {
				ByteArrayView containerAppBlob = tuple.getBlobBuffer(getContainerAppBlob());
				this.processAddContainer(containerName, containerAppBlob, tuple);
			}
			else {
				logger.error("Unknown ADD operation delivery type");
			}
		}
		else if(this.getOperation() == OperationEnum.MODIFY) {
			List<?> containerOptions = tuple.getList(getContainerOptions());		
			this.processModifyContainer(containerName, containerOptions, tuple);
		}
		else if(this.getOperation() == OperationEnum.REMOVE) {
			this.processRemoveContainer(containerName, tuple);
		}
		else {
			logger.error("Unknown operation " + this.getOperation());
			this.sendSuccessOutput(tuple);
		}			
	}

	/**
	 * Performs an add container from a file
	 * @param containerName container name
	 * @param containerAppFilename container app file name
	 * @param tuple
	 */
	private void processAddContainer(String containerName, String containerAppFilename, Tuple tuple) {
		Thread t = new Thread("container-operator-add") {
			public void run() {
				try {
					logger.debug("Loading container " + containerName + " from app file " + containerAppFilename);
					
			        byte[] containerApp = Util.readFileIntoBytes(new File(containerAppFilename));			        
			        ContainerOperator.this.addContainerToEngine(containerName, containerApp, tuple);
				}
				catch (StreamBaseException e) {
					logger.error(e.getLocalizedMessage());
					ContainerOperator.this.sendFailureOutput(tuple, "add", e.getLocalizedMessage());
				}
			}
		};
		
		t.start();
	}
	
	/**
	 * Performs an add container from an in-stream blob
	 * @param containerName container name
	 * @param containerAppBlob container app blob
	 * @param tuple
	 */
	private void processAddContainer(String containerName, ByteArrayView containerAppBlob, Tuple tuple ) {
		Thread t = new Thread("container-operator-add") {
			public void run() {
				try {
					logger.debug("Loading container " + containerName + " from blob");
					
			        byte[] containerApp = containerAppBlob.array();			        
			        ContainerOperator.this.addContainerToEngine(containerName, containerApp, tuple);
				}
				catch (StreamBaseException e) {
					logger.error(e.getLocalizedMessage());
					ContainerOperator.this.sendFailureOutput(tuple, "add", e.getLocalizedMessage());
				}
			}
		};
		
		t.start();
	}
	
	private void addContainerToEngine(String containerName, byte[] containerApp, Tuple tuple) throws StreamBaseException {
        // Start hack
        // This abysmal piece of code is to make up for a 10.3.0 regression on add container, still broken in 10.4.0
        /*String sbapp = new String(containerApp);
        String startDelim = "<type-metadata>";
        if(sbapp.indexOf(startDelim) > -1) {
	        String endDelim = "</type-metadata>";
	        StringBuffer sbappOut = new StringBuffer();
        	sbappOut.append(sbapp.substring(0, sbapp.indexOf(startDelim)));
        	sbappOut.append(sbapp.substring(sbapp.indexOf(endDelim) + endDelim.length()));
        	containerApp = sbappOut.toString().getBytes();
        }*/
        // End hack

        // Create args array to pass to the handler
        Object[] args = new Object[2];
        args[0] = containerName;
        args[1] = containerApp;
        
        (new ADD_CONTAINERHandler(ContainerOperator.this.server)).processJMX(args);
        
        ContainerOperator.this.sendSuccessOutput(tuple);
	}
	
	/**
	 * Performs a modify container
	 * @param containerName container name
	 * @param containerOptions container options
	 * @param tuple input tuple
	 * @throws StreamBaseException
	 */
	private void processModifyContainer(String containerName, List<?> containerOptions, Tuple tuple) throws StreamBaseException {
		Thread t = new Thread("container-operator-modify") {
			public void run() {
				try {
					logger.debug("Modifying container " + containerName + " app " + containerOptions);
					
			        String[] args = new String[containerOptions.size() + 1];
			        args[0] = containerName;
			        for (int i = 0; i < containerOptions.size(); i++) {
			            args[i + 1] = (String) containerOptions.get(i);
			        }
			
			        (new MODIFY_CONTAINERHandler(ContainerOperator.this.server)).processJMX(args);
					
			        ContainerOperator.this.sendSuccessOutput(tuple);
				}
				catch (StreamBaseException e) {
					logger.error(e.getLocalizedMessage());
					ContainerOperator.this.sendFailureOutput(tuple, "modify", e.getLocalizedMessage());
				}
			}
		};
		
		t.start();
	}
	
	/**
	 * Performs a remove container.
	 * @param containerName container to remove
	 * @param tuple input tuple
	 * @throws StreamBaseException
	 */
	private void processRemoveContainer(String containerName, Tuple tuple) throws StreamBaseException {
		Thread t = new Thread("container-operator-remove") {
			public void run() {
				try {
					logger.debug("Removing container " + containerName);
					
			        (new REMOVE_CONTAINERHandler(ContainerOperator.this.server)).processJMX(containerName);
			
			        ContainerOperator.this.sendSuccessOutput(tuple);
				}
				catch (StreamBaseException e) {
					logger.error(e.getLocalizedMessage());
					ContainerOperator.this.sendFailureOutput(tuple, "remove", e.getLocalizedMessage());
				}
			}
		};
		
		t.start();
	}
	
	/**
	 * Sends the output
	 * @param tuple input tuple
	 * @throws StreamBaseException
	 */
	private void sendSuccessOutput(Tuple tuple) {
		try {
			// create a new output tuple from the Schema at the port we are about to send to
			Tuple out = outputSchemas[0].createTuple();
	
			// Copy each field from the input
			for (int i = 0; i < out.getSchema().getFieldCount() - 1; ++i) {
				out.setField(i, tuple.getField(i));
			}
	
			// Create status
			Tuple status = statusSchema.createTuple();
			status.setBoolean("success", true);
			out.setTuple("status", status);
			
			sendOutput(0, out);
		}
		catch(StreamBaseException e) {
			logger.error(e.getMessage());
			this.sendFailureOutput(tuple, "sendSuccess", e.getMessage());
		}
	}
	
	
	/**
	 * Sends the output
	 * @param tuple input tuple
	 * @throws StreamBaseException
	 */
	private void sendFailureOutput(Tuple tuple, String reason, String message) {
		try {
			// create a new output tuple from the Schema at the port we are about to send to
			Tuple out = outputSchemas[0].createTuple();
	
			// Copy each field from the input
			for (int i = 0; i < out.getSchema().getFieldCount() - 1; ++i) {
				out.setField(i, tuple.getField(i));
			}
	
			// Create status
			Tuple status = statusSchema.createTuple();
			status.setBoolean("success", false);
			status.setString("reason", reason);
			status.setString("message", message);
			out.setTuple("status", status);
			
			sendOutput(0, out);
		}
		catch(StreamBaseException e) {
			logger.error(e.getMessage());
		}
	}
	
	
	/**
	 * Initialize the operator
	 */
	public void init() throws StreamBaseException {
		super.init();
		
		outputSchemas = new Schema[outputPorts];
		for (int i = 0; i < outputPorts; ++i) {
			outputSchemas[i] = getRuntimeOutputSchema(i);
		}
		
		// Get server reference
        this.server = (StreamBaseServer) this.getRuntimeEnvironment().xxgetInternalServerInstance();				
	}

	/**
	*  The shutdown method is called when the StreamBase server is in the process of shutting down.
	*/
	public void shutdown() {

	}

	/**
	 * 
	 * @param operation the operation
	 */
	public void setOperation(OperationEnum operation) {
		this.operation = operation;
		
		if(this.getOperation() == OperationEnum.MODIFY || this.getOperation() == OperationEnum.REMOVE) {
			this.setDelivery(null);
		}
	}

	/**
	 * 
	 * @return the operation
	 */
	public OperationEnum getOperation() {
		return this.operation;
	}

	/**
	 * 
	 * @param containerName container name
	 */
	public void setContainerName(String containerName) {
		this.containerName = containerName;
	}

	/**
	 * 
	 * @return container name
	 */
	public String getContainerName() {
		return this.containerName;
	}

	/**
	 * 
	 * @param delivery the delivery
	 */
	public void setDelivery(DeliveryEnum delivery) {
		this.delivery = delivery;
	}

	/**
	 * 
	 * @return the delivery
	 */
	public DeliveryEnum getDelivery() {
		return this.delivery;
	}

	/**
	 * 
	 * @param containerAppFilename container app filename
	 */
	public void setContainerAppFilename(String containerAppFilename) {
		this.containerAppFilename = containerAppFilename;
	}

	/**
	 * 
	 * @return container app blob
	 */
	public String getContainerAppBlob() {
		return this.containerAppBlob;
	}

	/**
	 * 
	 * @param containerAppBlob container app blob
	 */
	public void setContainerAppBlob(String containerAppBlob) {
		this.containerAppBlob = containerAppBlob;
	}

	/**
	 * 
	 * @return container app filename
	 */
	public String getContainerAppFilename() {
		return this.containerAppFilename;
	}

	/**
	 * 
	 * @param containerOptions container options
	 */
	public void setContainerOptions(String containerOptions) {
		this.containerOptions = containerOptions;
	}

	/**
	 * 
	 * @return container options
	 */
	public String getContainerOptions() {
		return this.containerOptions;
	}

	/**
	 * 
	 * @return enable operation
	 */
	public boolean shouldEnableOperation() {
		return true;
	}

	/**
	 * 
	 * @return enable delivery
	 */
	public boolean shouldEnableDelivery() {
		if(this.getOperation() == OperationEnum.ADD)
			return true;
		else
			return false;
	}

	/**
	 * 
	 * @return enable container name
	 */
	public boolean shouldEnableContainerName() {
		return true;
	}

	/**
	 * 
	 * @return enable container app filename
	 */
	public boolean shouldEnableContainerAppFilename() {
		if(this.getOperation() == OperationEnum.ADD && this.getDelivery() == DeliveryEnum.FILE)
			return true;
		else
			return false;
	}

	/**
	 * 
	 * @return enable container app filename
	 */
	public boolean shouldEnableContainerAppBlob() {
		if(this.getOperation() == OperationEnum.ADD && this.getDelivery() == DeliveryEnum.IN_STREAM_BLOB)
			return true;
		else
			return false;
	}

	/**
	 * 
	 * @return enable container options
	 */
	public boolean shouldEnableContainerOptions() {
		if(this.getOperation() == OperationEnum.MODIFY)
			return true;
		else
			return false;
	}

}
