/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.operators.container;

import java.net.URL;

import org.slf4j.Logger;

import com.streambase.sb.*;
import com.streambase.sb.operator.*;
import com.streambase.sb.runtime.Enqueuer;
import com.streambase.sb.runtime.StreamInfo;
import com.streambase.sb.sbd.StreamBaseServer;

public class LocalEnqueueOperator extends Operator implements Parameterizable {

	public static final long serialVersionUID = 1532077495726L;
	private Logger logger;

	// Properties
	private String target;
	private String payload;
	private Schema enqueueSchema;
	private String displayName = "Local Enqueue";

	// Local variables
	private int inputPorts = 1;
	private int outputPorts = 0;

	private StreamBaseServer server = null;
	
	/**
	 * Create new operator.
	 */
	public LocalEnqueueOperator() {
		super();
		
		logger = getLogger();
		setPortHints(inputPorts, outputPorts);
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setTarget("");
		setPayload("");

	}

	/**
	 * Typecheck
	 */
	public void typecheck() throws TypecheckException {
		// typecheck: require a specific number of input ports
		requireInputPortCount(inputPorts);
		setCaptureStrategy(CaptureTransformStrategy.NEST);

		// Validate start input
		Schema inputSchemaStart = getInputSchema(0);
		try {
			inputSchemaStart.getField(getTarget());
		} catch (TupleException e) {
			throw new PropertyTypecheckException("target", "Target output stream tuple not present in input");
		}		
		try {
			inputSchemaStart.getField(getPayload());			
		} catch (TupleException e) {
			throw new PropertyTypecheckException("payload", "Payload tuple not present in input");
		}	
		if(this.getEnqueueSchema() == null) {
			throw new PropertyTypecheckException("enqueueSchema", "Enqueue schema not present in input");
		}	
		else {
			try {
				if(!inputSchemaStart.getField(getPayload()).getSchema().equals(this.getEnqueueSchema())) {
					throw new PropertyTypecheckException("payload", "Payload schema does not match enqueue schema");
				}
			} catch (TupleException e) {
				throw new PropertyTypecheckException("payload", "Payload schema exception");
			}	
		}
	}

	/**
	 * Process tuple.
	 * @param inputPort the port
	 * @param tuple the tuple
	 */
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {

		String targetInput = tuple.getString(getTarget());
		Tuple payload = tuple.getTuple(getPayload());

		// Parse targetInput into container and stream names
		String containerName = targetInput.substring(0, targetInput.indexOf('.'));
		String streamName = targetInput.substring(targetInput.indexOf('.') + 1);

		// Get the stream info
		StreamInfo streamInfo = this.server.getNode().getContainerManager().getContainer(containerName).getModule().getMetaData().getStream(streamName, CaptureTransformStrategy.FLATTEN);
		Schema s = streamInfo.getProps().getSchema();
		
		// Create an output tuple
		Tuple out = s.createTuple();

		// Copy the input into the output. If the schemas don't match this is going to fail, possibly unpredictably.
        Schema.Field[] sourceFields = payload.getSchema().getFields();
        for (int i = 0; i < sourceFields.length; i++) {
            Schema.Field sourceField = sourceFields[i];
            out.setField(sourceField, payload.getField(sourceField));
        }    
        
		// Enqueue the target
		Enqueuer enq = streamInfo.getEnqueuer();
		enq.enqueue(out);

		logger.debug("Enqueued data to " + targetInput);
	}

	
	/**
	 * Initialize
	 * @throws StreamBaseException on exception
	 */
	public void init() throws StreamBaseException {
		super.init();	
		
        this.server = (StreamBaseServer) this.getRuntimeEnvironment().xxgetInternalServerInstance();
	}

	/**
	 * Shutdown.
	 */
	public void shutdown() {

	}

	@Override
	public URL getIconResource(IconKind iconType) {
		switch (iconType) {
		case CANVAS_OVERLAY_ICON:
			return getClass().getResource("icons/enqueue_48.png");
		case PALETTE_ICON_LARGE:
			return getClass().getResource("icons/enqueue_32.png");
		case PALETTE_ICON_SMALL:
			return getClass().getResource("icons/enqueue_16.png");
		}
		return null;
	}

	
	/**
	 * 
	 * @param target target to set
	 */
	public void setTarget(String target) {
		this.target = target;
	}

	/**
	 * 
	 * @return target
	 */
	public String getTarget() {
		return this.target;
	}

	/**
	 * 
	 * @param payload payload to set
	 */
	public void setPayload(String payload) {
		this.payload = payload;
	}

	/**
	 * 
	 * @return payload
	 */
	public String getPayload() {
		return this.payload;
	}

	/**
	 * 
	 * @param enqueueSchema the schema
	 */
	public void setEnqueueSchema(Schema enqueueSchema) {
		this.enqueueSchema = enqueueSchema;
	}

	/**
	 * 
	 * @return the schema
	 */
	public Schema getEnqueueSchema() {
		return this.enqueueSchema;
	}
	
	/**
	 * 
	 * @return enable target
	 */
	public boolean shouldEnableTarget() {
		return true;
	}

	/**
	 * 
	 * @return enable payload
	 */
	public boolean shouldEnablePayload() {
		return true;
	}

	/**
	 * 
	 * @return enable schema
	 */
	public boolean shouldEnableEnqueueSchema() {
		return true;
	}
	
}
