/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.operators.file.handler;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;

import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.tibco.accelerators.operators.file.FileListOperator;

public class FileListHandler extends Thread{
	private FileListOperator operator = null;
	private Logger logger = null;
	private Tuple inputTuple = null;

	public FileListHandler(Logger logger, FileListOperator operator, Tuple tuple) {
		this.logger = logger;
		this.operator = operator;
		this.inputTuple = tuple;
	}
	
	public void run() {
		final String listPath = this.getListPath();		
		final boolean searchSubDirs = this.operator.getSearchSubDirs();
		final String listFilter = this.getListFilter();
		
		// Create output
		Schema dataOutputSchema = this.operator.getCachedOutputSchema(0);
		Tuple dataOutputTuple = dataOutputSchema.createTuple();
		
		try {
			// Validate the list path exists
			this.validateListPathExists(listPath);
			
			String targetPattern = listFilter;
			if(targetPattern != null) {
				targetPattern = targetPattern.replace(".", "\\.");
				targetPattern = targetPattern.replace("?", ".?").replace("*", ".*");
			}
			final String regex = targetPattern;
			
			// Create the path file object
			File listPathFile = new File(listPath);
			List<File> fileList = this.processDirectory(listPathFile, regex, searchSubDirs);

			List<Tuple> fileTuples = new ArrayList<Tuple>();
			if(fileList != null) {
				for(int idx = 0; idx < fileList.size(); idx++) {
					File f = fileList.get(idx);
					Tuple fileOutputTuple = null;
					if(this.operator.getOneTuplePerFile() == true) {
						fileOutputTuple = dataOutputTuple;
					}
					else {
						Schema fileOutputSchema = dataOutputSchema.getField("files").getCompleteDataType().getElementType().getSchema();
						fileOutputTuple = fileOutputSchema.createTuple();
					}
					
					fileOutputTuple.setString("absolutePath", f.getAbsolutePath());
					fileOutputTuple.setString("canonicalPath", f.getCanonicalPath());
					fileOutputTuple.setString("name", f.getName());
					fileOutputTuple.setString("parent", f.getParent());
					fileOutputTuple.setString("path", f.getPath());
					fileOutputTuple.setTimestamp("lastModified", new Timestamp(new java.util.Date(f.lastModified())));
					fileOutputTuple.setLong("length", f.length());
					fileOutputTuple.setString("type", f.isDirectory() == true ? "directory" : f.isFile() == true ? "file" : "other");
	
					if(this.operator.getOneTuplePerFile() == true)
						this.sendDataTuple(fileOutputTuple);
					else
						fileTuples.add(fileOutputTuple);
				}
			}
			if(this.operator.getOneTuplePerFile() == false) {
				dataOutputTuple.setList("files", fileTuples);
				this.sendDataTuple(dataOutputTuple);
			}
		} catch (Exception ex) {
			//ex.printStackTrace();
			this.logger.error(ex.getMessage());
		}
	}

	private List<File> processDirectory(File listPathFile, final String regex, final boolean searchSubDirs) {
		final List<File> fileList = new ArrayList<File>();
		File[] fileArr = listPathFile.listFiles(new FileFilter() {				
			@Override
			public boolean accept(File file) {
				if(file.isFile() == true) {
					if(regex == null || regex.length() == 0) {
						return true;
					}
					else {
						return file.getName().matches(regex);
					}
				}
				else if(file.isDirectory() && searchSubDirs == true) {
					List<File> subDirFileList = processDirectory(file, regex, searchSubDirs);
					fileList.addAll(subDirFileList);
					return false;
				}
				else
					return false;
			}
		});
		
		if(fileArr != null)
			fileList.addAll(Arrays.asList(fileArr));
		return fileList;
	}
	
	private String getListPath() {
		// Setup the listPath
		String listPath = null;
		
		try {
			listPath = this.inputTuple.getString("listPath");
		} catch (TupleException e1) {
			listPath = this.operator.getListPath();
		}
		
		return listPath;
	}
	
	private String getListFilter() {
		// Setup the listFilter
		String listFilter = null;
		
		try {
			listFilter = this.inputTuple.getString("listFilter");
		} catch (TupleException e1) {
			listFilter = this.operator.getListFilter();
		}
		
		return listFilter;
	}
	
	private void validateListPathExists(String listPath) throws StreamBaseException {
		if(listPath == null || new File(listPath).exists() == false) {
			String message = "Error opening list path '" + listPath + "' not found";
			throw new StreamBaseException(message);
		}
		else if(new File(listPath).isDirectory() == false) {
			String message = "Error opening list path '" + listPath + "' is not a directory";
			throw new StreamBaseException(message);
		}
	}
	
	private void sendDataTuple(Tuple dataOutputTuple) throws StreamBaseException {
		// Append passthrough
		this.appendPassthrough(dataOutputTuple);
		
		// Send data tuple
		this.operator.sendOutput(0, dataOutputTuple);
	}
		
	private void appendPassthrough(Tuple dataOutputTuple) throws TupleException {
		if(this.operator.getPassthroughFields() == true) {
			dataOutputTuple.setField("passthrough", this.inputTuple.getField("passthrough"));
		}
	}
	
}
