/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.operators.file.handler;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;

import com.streambase.sb.ByteArrayView;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.TupleJSONUtil;
import com.streambase.sb.operator.ResourceNotFoundException;
import com.streambase.sb.util.CSV;
import com.streambase.sb.util.CSV.FormatInfo;
import com.tibco.accelerators.operators.file.FileReaderOperator;
import com.tibco.accelerators.operators.file.parse.CsvToTupleParser;
import com.tibco.accelerators.operators.file.parse.CsvToTupleParser.LineParseResult;

public class FileReaderHandler extends Thread {
	private FileReaderOperator operator = null;
	private Logger logger = null;
	private Tuple inputTuple = null;

	public FileReaderHandler(Logger logger, FileReaderOperator operator, Tuple tuple) {
		this.logger = logger;
		this.operator = operator;
		this.inputTuple = tuple;
	}

	public void run() {
		// Setup the filename
		String filename = this.getFilename();
		this.sendEventTuple("Read", "Start", filename, "Start processing");

		try {
			// Validate the file exists
			this.validateFileExists(filename);

			// Read the file contents
			byte[] fileContentBytes = this.readFileContents(filename);
			
			// Setup the encoding
			String encoding = this.operator.getCharacterSet();
			if (encoding == null || encoding.length() == 0)
				encoding = Charset.defaultCharset().name();
			
			String fileContent = new String(fileContentBytes, encoding);

			// Choose send action based on Parse As
			if (this.operator.getParseAs().equals("None")) {
				this.parseForNone(filename, fileContentBytes, fileContent);
			} else if (this.operator.getParseAs().equals("Line")) {
				this.parseForLine(filename, fileContent);
			} else if (this.operator.getParseAs().equals("CSV")) {
				this.parseForCsv(filename, fileContent);
			} else if (this.operator.getParseAs().equals("JSON")) {
				this.parseForJson(filename, fileContent);
			}

			this.sendEventTuple("Read", "Finish", filename,
					"Completed processing");

		} catch (Exception ex) {
			//ex.printStackTrace();
			this.logger.error(ex.getMessage());
			this.sendEventTuple("Read", "Error", filename, ex.getMessage());
		}
	}

	private String getFilename() {
		// Setup the filename
		String filename = null;

		try {
			filename = this.inputTuple.getString("filename");
		} catch (TupleException e1) {
			filename = this.operator.getFilename();
		}

		return filename;
	}

	private void validateFileExists(String filename) throws StreamBaseException {
		boolean valid = true;
		if(filename == null) {
			valid = false;
		}
		else if(this.operator.getReadAsResource() == true) {
			try {
				this.operator.getResourceFile(filename);
			}
			catch(ResourceNotFoundException rnfex) {
				valid = false;
			}
		}
		else if (new File(filename).exists() == false) {
			valid = false;
		}
		
		if(valid == false) {
			String message = "Error opening file '" + filename + "' not found";
			throw new StreamBaseException(message);
		}
	}

	private byte[] readFileContents(String filename) throws IOException {
		// Send event tuple
		this.sendEventTuple("Read", "Open", filename, "File opened");

		// Get the file reference
		File file;
		if(this.operator.getReadAsResource() == true) {
			try {
				file = this.operator.getResourceFile(filename);
			} 
			catch (StreamBaseException e) {
				throw new IOException(e.getMessage(), e);
			}
		}
		else {
			file = new File(filename);
		}
		
		// Load the file contents as bytes
		InputStream inputStream = new FileInputStream(file);
		ByteArrayOutputStream result = new ByteArrayOutputStream();
		byte[] buffer = new byte[1024];
		int length;
		while ((length = inputStream.read(buffer)) != -1) {
			result.write(buffer, 0, length);
		}
		inputStream.close();
		result.close();
		
		// Send event tuple
		this.sendEventTuple("Read", "Close", filename, "File closed");

		// Return buffer
		return result.toByteArray();
	}

	private void sendEventTuple(String type, String action, String object,
			String message) {
		if (this.operator.getEventPort() == true) {
			Schema eventOutputSchema = this.operator.getCachedOutputSchema(1);
			Tuple eventOutputTuple = eventOutputSchema.createTuple();
			try {
				eventOutputTuple.setString("type", type);
				eventOutputTuple.setString("action", action);
				eventOutputTuple.setString("object", object);
				eventOutputTuple.setString("message", message);

				this.operator.sendOutput(1, eventOutputTuple);
			} catch (StreamBaseException se) {
				se.printStackTrace();
			}
		}
	}

	private void sendDataTuple(Tuple dataOutputTuple)
			throws StreamBaseException {
		// Append passthrough
		this.appendPassthrough(dataOutputTuple);

		// Send data tuple
		this.operator.sendOutput(0, dataOutputTuple);
	}

	private void parseForNone(String filename, byte[] fileContentAsBytes,
			String fileContent) throws StreamBaseException {
		// Create the data output tuple
		Schema dataOutputSchema = this.operator.getCachedOutputSchema(0);
		Tuple dataOutputTuple = dataOutputSchema.createTuple();
		dataOutputTuple.setString("filename", filename);

		if (this.operator.getReadAsBinary() == true) {
			ByteArrayView byteArrayView = ByteArrayView
					.makeCopiedView(fileContentAsBytes);
			dataOutputTuple.setBlobBuffer("content", byteArrayView);
		} else {
			dataOutputTuple.setString("content", fileContent);
		}

		this.sendDataTuple(dataOutputTuple);
	}

	private void parseForLine(String filename, String fileContent)
			throws StreamBaseException {
		// Create the data output tuple
		Schema dataOutputSchema = this.operator.getCachedOutputSchema(0);

		// Create the parser
		CsvToTupleParser parser = new CsvToTupleParser(this.logger);
		LineParseResult result = parser.parseStringToLines(
				this.operator.getLineSeparator(), this.operator.getHeaderTypeAction(),
				fileContent);

		List<String> lines = result.getLines();
		if (this.operator.getOneTuplePerLine() == true) {
			for (int idx = 0; idx < lines.size(); idx++) {
				Tuple dataOutputTuple = dataOutputSchema.createTuple();
				dataOutputTuple.setString("filename", filename);
				dataOutputTuple.setString("line", lines.get(idx));

				this.sendDataTuple(dataOutputTuple);
			}
		} else {
			Tuple dataOutputTuple = dataOutputSchema.createTuple();
			dataOutputTuple.setString("filename", filename);
			dataOutputTuple.setList(this.operator.getListTupleName(), lines);

			this.sendDataTuple(dataOutputTuple);
		}
	}

	private void parseForCsv(String filename, String fileContent)
			throws StreamBaseException {
		// Create the data output tuple
		Schema dataOutputSchema = this.operator.getCachedOutputSchema(0);
		Schema schemaOut = this.operator.getSchemaOut();

		// Create the parser
		CsvToTupleParser parser = new CsvToTupleParser(this.logger);
		LineParseResult result = parser.parseStringToLines(
				this.operator.getLineSeparator(), this.operator.getHeaderTypeAction(),
				fileContent);
		List<Tuple> tuples = new ArrayList<Tuple>();

		// Process the lines
		List<String> lines = result.getLines();
		for (int idx = 0; idx < lines.size(); idx++) {
			// Create tuple and parse the CSV
			Tuple tuple = schemaOut.createTuple();
			FormatInfo format = CSV.DEFAULT_FORMAT;
			if(this.operator.getEmptyAsNull() == true) {
				format = new CSV.FormatInfo(CSV.DEFAULT_QUOTE, CSV.DEFAULT_DELIM, "", null);
			}
			CSV.readCSVLine(lines.get(idx), format, tuple);
			tuples.add(tuple);
		}
		
		if (this.operator.getOneTuplePerLine() == true) {
			for (int idx = 0; idx < tuples.size(); idx++) {
				Tuple tuple = this.schemaOutTupleToDataOutputTuple(tuples.get(idx), dataOutputSchema, schemaOut);
				this.sendDataTuple(tuple);
			}
		} else {
			Tuple dataOutputTuple = dataOutputSchema.createTuple();
			dataOutputTuple.setList(this.operator.getListTupleName(), tuples);
			this.sendDataTuple(dataOutputTuple);
		}
	}

	private void parseForJson(String filename, String fileContent)
			throws StreamBaseException {
		// Create the data output tuple
		Schema dataOutputSchema = this.operator.getCachedOutputSchema(0);
		Schema schemaOut = this.operator.getSchemaOut();
		
		// parse the tuples
		List<Tuple> tuples = TupleJSONUtil.getTuplesFromJSON(schemaOut, fileContent);
		
		// Handle send
		for (int idx = 0; idx < tuples.size(); idx++) {
			Tuple tuple = this.schemaOutTupleToDataOutputTuple(tuples.get(idx), dataOutputSchema, schemaOut);
			this.sendDataTuple(tuple);
		}
	}

	private Tuple schemaOutTupleToDataOutputTuple(Tuple schemaOutTuple, Schema dataOutputSchema, Schema schemaOut) throws TupleException {
		Tuple tuple = dataOutputSchema.createTuple();
		for(int idx = 0; idx < schemaOut.getFieldCount(); idx++) {
			tuple.setField(schemaOut.getField(idx).getName(), schemaOutTuple.getField(idx));
		}
		return tuple;
	}
	
	private void appendPassthrough(Tuple dataOutputTuple) throws TupleException {
		if (this.operator.getPassthroughFields() == true) {
			dataOutputTuple.setField("passthrough",
					this.inputTuple.getField("passthrough"));
		}
	}
}
