/*
* Copyright © 2020. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

package com.tibco.accelerators.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;

import org.jose4j.json.internal.json_simple.JSONArray;
import org.jose4j.json.internal.json_simple.JSONObject;
import org.jose4j.json.internal.json_simple.JSONValue;
import org.jose4j.json.internal.json_simple.parser.ParseException;

@SuppressWarnings("unchecked") 
public class SwaggerToSchema {
	private JSONObject rootObj;
	private Map<String, JSONObject> definitionMap = new HashMap<String, JSONObject>();
	
	public SwaggerToSchema(String payload) throws ParseException {
		// Parse to root object
		this.rootObj = (JSONObject) JSONValue.parseWithException(payload);
	}
	
	public String process() {
		StringBuffer sb = new StringBuffer();
		sb.append(this.header());
		
		// Get the definitions object
		JSONObject definitionsObj = (JSONObject) this.rootObj.get("definitions");
		
		// Iterate over the child objects in the definitions
		Set<Map.Entry<String,Object>> definitionsSet = definitionsObj.entrySet();
		for(Map.Entry<String, Object> thisDefinition : definitionsSet) {
			String definitionName = thisDefinition.getKey();
			JSONObject definitionObj = (JSONObject) thisDefinition.getValue();
			definitionMap.put(definitionName, definitionObj);
		}
		
		for(Map.Entry<String, Object> thisDefinition : definitionsSet) {
			String definitionName = thisDefinition.getKey();
			JSONObject definitionObj = (JSONObject) thisDefinition.getValue();
			sb.append(processDefinition(definitionName, definitionObj));
		}
		
		sb.append(this.footer());
		
		System.out.println("DONE");
		return sb.toString();
	}
	
	private String header() {
		StringBuffer sb = new StringBuffer();
		sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n");
		sb.append("<application version=\"10.6.1_0a0fdf7f8f3d4f25851d53e0e55c97ce2ece3d22\">\n");
		sb.append("<annotations>\n");
		sb.append("<annotation name=\"hygienic\"/>\n");
		sb.append("</annotations>\n");
		sb.append("<type-metadata>\n");
		sb.append("<param name=\"type\" value=\"interface\"/>\n");
		sb.append("<param name=\"fully-qualified-name\" value=\"com.tibco.accelerators.openbanking.common.resources.schemas.CDR\"/>\n");
		sb.append("</type-metadata>\n");
		sb.append("<memory-model-settings/>\n");
		sb.append("<named-schemas>\n");
		
		return sb.toString();
	}
	
	private String footer() {
		StringBuffer sb = new StringBuffer();
		sb.append("</named-schemas>\n");
		sb.append("<dynamic-variables/>\n");
		sb.append("</application>\n");
		
		return sb.toString();
	}
	
	private String processDefinition(String definitionName, JSONObject definitionObj) {
		//if(!definitionName.equals("RequestAccountIds"))
		//	return null;
		
		// Create buffer to hold the XML
		StringBuffer sb = new StringBuffer();
		
		// Start the schema
		sb.append("<schema name=\"" + definitionName + "\">\n");
		
		// See if there is an allOf object
		JSONArray allOfArr = (JSONArray) definitionObj.get("allOf");
		JSONObject propertiesObj = (JSONObject) definitionObj.get("properties");
		
		if(allOfArr != null) {			
			List<String> extendSchemas = new ArrayList<String>();
			List<String> fieldSchemas = new ArrayList<String>();

			for(int idx = 0; idx < allOfArr.size(); idx++) {
				JSONObject thisObject = (JSONObject) allOfArr.get(idx);
				
				String thisObjectRef = (String) thisObject.get("$ref");
				JSONObject thisPropertiesObj = (JSONObject) thisObject.get("properties");
				if(thisObjectRef != null) {
					extendSchemas.add(this.getType(thisObjectRef));
				}
				else if(thisPropertiesObj != null) {
					fieldSchemas.add(processProperties(thisPropertiesObj));
				}
				else {
					System.out.println("[WARN] Don't know what to do with allOf for " + definitionName);
				}
			}
			
			if(extendSchemas.size() > 0) {					
				sb.append("<parent-schemas>\n");
				for(String thisName : extendSchemas)
					sb.append("<parent-schema name=\"" + thisName + "\"/>\n");
				sb.append("</parent-schemas>\n");
			}
			
			for(String thisField : fieldSchemas)
				sb.append(thisField);
		}
		else if(propertiesObj != null){
			sb.append(processProperties(propertiesObj));
		}
		else {
			System.out.println("[WARN] Don't know what to do with " + definitionName);
		}
		
		sb.append("\n</schema>\n");

		return sb.toString();
	}
	
	private String processProperties(JSONObject propertiesObj) {
		if(propertiesObj == null)
			return "";
		StringBuffer sb = new StringBuffer();

		Set<Map.Entry<String,Object>> propertiesSet = propertiesObj.entrySet();
		for(Map.Entry<String, Object> thisProperty : propertiesSet) {
			String propertyName = thisProperty.getKey();
			JSONObject propertyObj = (JSONObject) thisProperty.getValue();
			sb.append(processProperty(propertyName, propertyObj));
		}
		
		return sb.toString();
	}
	
	
	/*
	 * Process a single property
	 */
	private String processProperty(String propertyName, JSONObject propertyObj) {
		String refString = (String) propertyObj.get("$ref");

		// Is there a $ref
		if(refString != null) {
			JSONObject thisRefObj = this.definitionMap.get(this.getType(refString));
			
			// What is the ref type
			String thisRefObjType = (String) thisRefObj.get("type");
			
			// Is it a tuple?
			if(this.isRefObjectTuple(thisRefObjType, thisRefObj))
	            return "<field name=\"" + propertyName + "\" " +
						"type=\"tuple\" " + 
						"value=\"" + this.getType(refString) + "\"/>\n";
			// Otherwise is it a string?
			else if(thisRefObjType != null && thisRefObjType.equals("string"))
	            return "<field name=\"" + propertyName + "\" " +
						"type=\"string\"/>\n";
			// Otherwise log an error for now, other types might work as a string
			else {
				System.out.println("[WARN] UnknownRefObjectType: " + thisRefObjType + " refString " + refString + " property " + propertyName);
				return "";
			}
		}
		// Not a ref
		else {
			// Get the type
			String typeString = (String) propertyObj.get("type");
			
			// Handle the type
			if(typeString.equals("string"))
				return "<field name=\"" + propertyName + "\" type=\"string\"/>\n";
			else if(typeString.equals("boolean"))
				return "<field name=\"" + propertyName + "\" type=\"bool\"/>\n";
			else if(typeString.equals("integer"))
				return "<field name=\"" + propertyName + "\" type=\"int\"/>\n";
			else if(typeString.equals("number"))
				return "<field name=\"" + propertyName + "\" type=\"double\"/>\n";
			else if(typeString.equals("array")) {
				JSONObject itemsObj = (JSONObject) propertyObj.get("items");
				String itemsRefString = (String) itemsObj.get("$ref");
				String itemsTypeString = (String) itemsObj.get("type");
				
				// Is it a reference
				if(itemsRefString != null) {
					return "<field name=\"" + propertyName + "\" type=\"list\">" + 
								"<element-type type=\"tuple\" value=\"" + this.getType(itemsRefString) + "\"/>" + 
						   "</field>\n";
				}
				// Is it an object
				else if(itemsTypeString.equals("object")) {
					JSONObject thisObjectProperties = (JSONObject) propertyObj.get("properties");
					return "<field name=\"" + propertyName + "\" type=\"tuple\">" +
							"<schema>" +
								processProperties(thisObjectProperties) +
							"</schema>" +
							"</field>";				
				}
				// Otherwise log
				else {
					return "<field name=\"" + propertyName + "\" type=\"list\">" + 
								"<element-type type=\"" + itemsTypeString + "\"/>" + 
							"</field>\n";
				}
			}
			else if(typeString.equals("object")) {
				JSONObject thisObjectProperties = (JSONObject) propertyObj.get("properties");
				return "<field name=\"" + propertyName + "\" type=\"tuple\">" +
						"<schema>" +
							processProperties(thisObjectProperties) +
						"</schema>" +
						"</field>";				
			}
			else {
				System.out.println("[WARN] UnknownType: " + typeString + " property " + propertyName);
				return "";
			}
		}
	}
	
	/*
	 * Determine if a refObject is a tuple or something else
	 */
	private boolean isRefObjectTuple(String refObjType, JSONObject refObj) {
		boolean isTuple = false;
		JSONObject thisRefObjProperties = (JSONObject) refObj.get("properties");
		JSONArray thisRefObjAllOf = (JSONArray) refObj.get("allOf");
		if(thisRefObjProperties != null || thisRefObjAllOf != null || refObjType != null && refObjType.equals("object"))
			isTuple = true;
		return isTuple;
	}
	
	/*
	 * Extract the type from a $ref string
	 */
	private String getType(String refString) {
		return refString.substring(refString.lastIndexOf('/') + 1);
	}

	private static String readFile(String filename) throws FileNotFoundException {
		Scanner reader = new Scanner(new File(filename));
		StringBuffer sb = new StringBuffer();
		while(reader.hasNext()) {
			sb.append(reader.nextLine());
		}
		reader.close();

		return sb.toString();
	}
	
	private static void writeFile(String filename, String content) throws IOException {
		FileWriter writer = new FileWriter(filename);
		writer.write(content);
		writer.close();
	}
	
	
	public static void main(String[] args) throws ParseException, IOException {
		String sourceFilename = "C:\\Users\\Administrator\\Desktop\\cds_full.json";
		String targetFilename = "C:\\Users\\Administrator\\Desktop\\CDR.sbint";
		
		String swaggerJson = SwaggerToSchema.readFile(sourceFilename);
		
		SwaggerToSchema ss = new SwaggerToSchema(swaggerJson);
		String schemaXml = ss.process();
		SwaggerToSchema.writeFile(targetFilename, schemaXml);
		
		System.out.println(schemaXml);
	}
}
